@@ -159,11 +159,7 @@ pub async fn insert<D: Serialize, P: Into<PartitionKeys>, C: ToString>(
 
 /// Upsert a document into the cosmos database and returning an etag from the response if
 /// successful
-pub async fn upsert<
-    D: Serialize,
-    P: Into<PartitionKeys>,
-    C: ToString,
->(
+pub async fn upsert<D: Serialize, P: Into<PartitionKeys>, C: ToString>(
     collection_name: C,
     pk: P,
     document: &D,
@@ -173,12 +169,7 @@ pub async fn upsert<
 }
 
 /// Returns a specific document from the cosmos DB together with a corresponding etag
-pub async fn get<
-    D: DeserializeOwned,
-    P: Into<PartitionKeys>,
-    C: ToString,
-    S: ToString,
->(
+pub async fn get<D: DeserializeOwned, P: Into<PartitionKeys>, C: ToString, S: ToString>(
     collection_name: C,
     pk: P,
     document_id: S,
@@ -221,18 +212,18 @@ pub async fn get<
 /// If the transform closure fails then no insertion is performed and the error it fails with
 /// is returned
 pub async fn modify_async_get_old<
-D: Serialize + DeserializeOwned + Clone,
-P: Into<PartitionKeys>,
-F: Fn(D) -> Fut,
-C: ToString,
-S: ToString,
-Fut: futures::Future<Output = Result<D, warp::Rejection>>,
+    D: Serialize + DeserializeOwned + Clone,
+    P: Into<PartitionKeys>,
+    F: Fn(D) -> Fut,
+    C: ToString,
+    S: ToString,
+    Fut: futures::Future<Output = Result<D, warp::Rejection>>,
 >(
     collection_name: C,
     pk: P,
     document_id: S,
     transform: F,
-    ) -> Result<(D, D, String), CosmosError> {
+) -> Result<(D, D, String), CosmosError> {
     let collection_name = collection_name.to_string();
     let document_id = document_id.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&COSMOS_MASTER_KEY)
@@ -240,8 +231,7 @@ Fut: futures::Future<Output = Result<D, warp::Rejection>>,
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client =
-        CosmosClient::new(http_client, COSMOS_ACCOUNT.to_string(), authorization_token);
+    let client = CosmosClient::new(http_client, COSMOS_ACCOUNT.to_string(), authorization_token);
     let database_client = client.into_database_client(&*COSMOS_DATABASE);
     let collection_client = database_client.into_collection_client(collection_name);
     let document_client = collection_client
@@ -254,27 +244,27 @@ Fut: futures::Future<Output = Result<D, warp::Rejection>>,
             .await
             .map_err(|e| {
                 RetryLoopError::Permanent(new_cosmos_error(format!(
-                            "Could not get document {:?}",
-                            e
-                            )))
+                    "Could not get document {:?}",
+                    e
+                )))
             })? {
-                azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
-                azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
-                    return Err(RetryLoopError::Permanent(new_cosmos_error_kind(
-                                format!("Document not found: {:?}", resp),
-                                CosmosErrorKind::NotFound,
-                                )));
-                }
-            };
+            azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
+            azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
+                return Err(RetryLoopError::Permanent(new_cosmos_error_kind(
+                    format!("Document not found: {:?}", resp),
+                    CosmosErrorKind::NotFound,
+                )));
+            }
+        };
         let doc: D = resp.document.document;
         let old_doc = doc.clone();
 
         // Perform changes to the document
         let doc = transform(doc).await.map_err(|e| {
             RetryLoopError::Permanent(new_cosmos_error_kind(
-                    format!("Modification not possible: {:?}", e),
-                    CosmosErrorKind::InternalError,
-                    ))
+                format!("Modification not possible: {:?}", e),
+                CosmosErrorKind::InternalError,
+            ))
         })?;
         let c = collection_client
             .create_document()
@@ -292,14 +282,14 @@ Fut: futures::Future<Output = Result<D, warp::Rejection>>,
                 //means we need to retry the entire read/write block
                 if s.contains("412") {
                     return Err(RetryLoopError::Transient(new_cosmos_error_kind(
-                                format!("Cosmos db error: {:?}", s),
-                                CosmosErrorKind::PreconditionFailed,
-                                )));
+                        format!("Cosmos db error: {:?}", s),
+                        CosmosErrorKind::PreconditionFailed,
+                    )));
                 } else {
                     return Err(RetryLoopError::Permanent(new_cosmos_error(format!(
-                                    "Cosmos db error: {:?}",
-                                    s
-                                    ))));
+                        "Cosmos db error: {:?}",
+                        s
+                    ))));
                 }
             }
         };
@@ -415,17 +405,17 @@ impl<D> ModifyReturn<D> {
 /// Modifies a document in cosmos by applying `transform` closure on the existing document and
 /// either inserting the returned value or just returning it to the caller
 pub async fn maybe_modify<
-D: Serialize + DeserializeOwned + std::fmt::Debug,
-P: Into<PartitionKeys>,
-F: Fn(D) -> Result<ModifyReturn<D>, warp::Rejection>,
-C: ToString,
-S: ToString,
+    D: Serialize + DeserializeOwned + std::fmt::Debug,
+    P: Into<PartitionKeys>,
+    F: Fn(D) -> Result<ModifyReturn<D>, warp::Rejection>,
+    C: ToString,
+    S: ToString,
 >(
     collection_name: C,
     pk: P,
     document_id: S,
     transform: F,
-    ) -> Result<ModifyReturn<D>, CosmosError> {
+) -> Result<ModifyReturn<D>, CosmosError> {
     let collection_name = collection_name.to_string();
     let document_id = document_id.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&COSMOS_MASTER_KEY)
@@ -433,8 +423,7 @@ S: ToString,
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client =
-        CosmosClient::new(http_client, COSMOS_ACCOUNT.to_string(), authorization_token);
+    let client = CosmosClient::new(http_client, COSMOS_ACCOUNT.to_string(), authorization_token);
     let database_client = client.into_database_client(&*COSMOS_DATABASE);
     let collection_client = database_client.into_collection_client(collection_name);
     let document_client = collection_client
@@ -446,23 +435,23 @@ S: ToString,
             .execute::<D>()
             .await
             .map_err(|e| RetryLoopError::Permanent(new_cosmos_error(format!("{:?}", e))))?
-            {
-                azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
-                azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
-                    return Err(RetryLoopError::Permanent(new_cosmos_error_kind(
-                                format!("Document not found: {:?}", resp),
-                                CosmosErrorKind::NotFound,
-                                )));
-                }
-            };
+        {
+            azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
+            azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
+                return Err(RetryLoopError::Permanent(new_cosmos_error_kind(
+                    format!("Document not found: {:?}", resp),
+                    CosmosErrorKind::NotFound,
+                )));
+            }
+        };
         let doc: D = resp.document.document;
 
         // Perform changes to the document
         let doc = transform(doc).map_err(|e| {
             RetryLoopError::Permanent(new_cosmos_error_kind(
-                    format!("Modification error: {:?}", e),
-                    CosmosErrorKind::InternalError,
-                    ))
+                format!("Modification error: {:?}", e),
+                CosmosErrorKind::InternalError,
+            ))
         })?;
 
         let doc = match doc {
@@ -477,21 +466,23 @@ S: ToString,
             .if_match_condition(IfMatchCondition::Match(&resp.etag));
 
         match c.execute(&Document::new(&doc)).await {
-            Ok(_) => return Result::Ok::<_, RetryLoopError<CosmosError>>(ModifyReturn::Replace(doc)),
+            Ok(_) => {
+                return Result::Ok::<_, RetryLoopError<CosmosError>>(ModifyReturn::Replace(doc))
+            }
             Err(err) => {
                 let s = err.to_string();
                 //NOTE: 412 means the document has been edited between read and write so it
                 //means we need to retry the entire read/write block
                 if s.contains("412") {
                     return Err(RetryLoopError::Transient(new_cosmos_error_kind(
-                                format!("Cosmos db error: {:?}", s),
-                                CosmosErrorKind::PreconditionFailed,
-                                )));
+                        format!("Cosmos db error: {:?}", s),
+                        CosmosErrorKind::PreconditionFailed,
+                    )));
                 } else {
                     return Err(RetryLoopError::Transient(new_cosmos_error(format!(
-                                    "Cosmos db error: {:?}",
-                                    s
-                                    ))));
+                        "Cosmos db error: {:?}",
+                        s
+                    ))));
                 }
             }
         };
@@ -503,17 +494,17 @@ S: ToString,
 /// Modify in cosmos but does not retry on failure which is useful when one needs to pass in a
 /// `FnOnce` closure
 pub async fn modify_no_retry<
-D: Serialize + DeserializeOwned + std::fmt::Debug,
-P: Into<PartitionKeys>,
-F: FnOnce(D) -> Result<D, warp::Rejection>,
-C: ToString,
-S: ToString,
+    D: Serialize + DeserializeOwned + std::fmt::Debug,
+    P: Into<PartitionKeys>,
+    F: FnOnce(D) -> Result<D, warp::Rejection>,
+    C: ToString,
+    S: ToString,
 >(
     collection_name: C,
     pk: P,
     document_id: S,
     transform: F,
-    ) -> Result<D, CosmosError> {
+) -> Result<D, CosmosError> {
     let collection_name = collection_name.to_string();
     let document_id = document_id.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&COSMOS_MASTER_KEY)
@@ -521,8 +512,7 @@ S: ToString,
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client =
-        CosmosClient::new(http_client, COSMOS_ACCOUNT.to_string(), authorization_token);
+    let client = CosmosClient::new(http_client, COSMOS_ACCOUNT.to_string(), authorization_token);
     let database_client = client.into_database_client(&*COSMOS_DATABASE);
     let collection_client = database_client.into_collection_client(collection_name);
     let document_client = collection_client
@@ -533,15 +523,15 @@ S: ToString,
         .execute::<D>()
         .await
         .map_err(|e| new_cosmos_error(format!("{:?}", e)))?
-        {
-            azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
-            azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
-                return Err(new_cosmos_error_kind(
-                        format!("Document not found: {:?}", resp),
-                        CosmosErrorKind::NotFound,
-                        ));
-            }
-        };
+    {
+        azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
+        azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
+            return Err(new_cosmos_error_kind(
+                format!("Document not found: {:?}", resp),
+                CosmosErrorKind::NotFound,
+            ));
+        }
+    };
     let doc: D = resp.document.document;
 
     // Perform changes to the document
@@ -549,7 +539,7 @@ S: ToString,
         new_cosmos_error_kind(
             format!("Modification error: {:?}", e),
             CosmosErrorKind::InternalError,
-            )
+        )
     })?;
     let c = collection_client
         .create_document()
@@ -566,7 +556,6 @@ S: ToString,
     Ok(doc)
 }
 
-
 /// Modifies a document in cosmos by applying `transform` closure on the existing document and then
 /// inserting the returned document and returning the transformed document if successful
 /// If the transform closure fails then no insertion is performed and the error it fails with
@@ -621,11 +610,7 @@ pub async fn delete<C: ToString, S: ToString, P: Into<PartitionKeys>>(
     Ok(())
 }
 
-pub async fn query_crosspartition_etag<
-    D: DeserializeOwned,
-    P: Into<PartitionKeys>,
-    C: ToString,
->(
+pub async fn query_crosspartition_etag<D: DeserializeOwned, P: Into<PartitionKeys>, C: ToString>(
     collection_name: C,
     pk: P,
     query: String,
@@ -634,66 +619,72 @@ pub async fn query_crosspartition_etag<
 ) -> Result<Vec<(D, String)>, CosmosError> {
     let collection_name = collection_name.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&COSMOS_MASTER_KEY)
-        .map_err(into_cosmos_error("Could not get authorization token"))?;
+        .map_err(into_cosmos_error("Could not get authorization token."))?;
     let pk: PartitionKeys = pk.into();
-
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
     let client = CosmosClient::new(http_client, COSMOS_ACCOUNT.to_string(), authorization_token);
     let database_client = client.into_database_client(&*COSMOS_DATABASE);
-    let collection_client = database_client.into_collection_client(collection_name);
-    let query = Query::new(&query);
-    let mut query_documents_builder = collection_client
-        .query_documents()
-        .max_item_count(max_count);
-    if cross_partition {
-        query_documents_builder = query_documents_builder.query_cross_partition(true);
-    } else {
-        query_documents_builder = query_documents_builder.partition_keys(&pk);
+    let collection_client = database_client.into_collection_client(collection_name.clone());
+    let mut documents: Vec<(D, String)> = vec![];
+    let mut continuation_token = String::from("");
+    loop {
+        let q = Query::new(&query);
+        let mut query_documents_builder = collection_client
+            .query_documents()
+            .max_item_count(max_count);
+        if cross_partition {
+            query_documents_builder = query_documents_builder.query_cross_partition(true);
+        } else {
+            query_documents_builder = query_documents_builder.partition_keys(&pk);
+        }
+        if continuation_token != "" {
+            query_documents_builder =
+                query_documents_builder.continuation(continuation_token.as_str());
+        }
+        let query_documents_response =
+            query_documents_builder
+                .execute::<D, _>(q)
+                .await
+                .map_err(into_cosmos_error(format!(
+                    "Could not get query documents: {} {} {:?}.",
+                    collection_name, query, pk
+                )))?;
+        let query_documents_response =
+            query_documents_response
+                .into_documents()
+                .map_err(into_cosmos_error(
+                    "Could not get cosmos db query document response.",
+                ))?;
+        let mut fetched_documents: Vec<(D, String)> = query_documents_response
+            .results
+            .into_iter()
+            .map(|document| {
+                (
+                    document.result,
+                    document.document_attributes.etag().to_string(),
+                )
+            })
+            .collect();
+        documents.append(&mut fetched_documents);
+        continuation_token = match query_documents_response.continuation_token {
+            Some(token) => token.clone(),
+            None => {
+                break;
+            }
+        };
     }
-    let query_documents_response = query_documents_builder
-        .execute::<D, _>(query)
-        .await
-        .map_err(into_cosmos_error("Could not get query documents"))?;
-
-    let query_documents_response =
-        query_documents_response
-            .into_documents()
-            .map_err(into_cosmos_error(
-                "Could not get cosmos db query document response",
-            ))?;
-
-    let documents: Vec<(D, String)> = query_documents_response
-        .results
-        .into_iter()
-        .map(|document| {
-            (
-                document.result,
-                document.document_attributes.etag().to_string(),
-            )
-        })
-        .collect();
     Ok(documents)
 }
 
-pub async fn query_crosspartition<
-    D: DeserializeOwned,
-    P: Into<PartitionKeys>,
-    C: ToString,
->(
+pub async fn query_crosspartition<D: DeserializeOwned, P: Into<PartitionKeys>, C: ToString>(
     collection_name: C,
     pk: P,
     query: String,
     max_count: i32,
     cross_partition: bool,
 ) -> Result<Vec<D>, CosmosError> {
-    let v = query_crosspartition_etag(
-        collection_name,
-        pk,
-        query,
-        max_count,
-        cross_partition,
-    )
-    .await?;
+    let v =
+        query_crosspartition_etag(collection_name, pk, query, max_count, cross_partition).await?;
     Ok(v.into_iter().map(|(d, _)| d).collect())
 }
 
@@ -723,7 +714,7 @@ pub async fn upload_image(mut f: FormData) -> Result<String, CosmosError> {
                                         let ext = String::from(n);
                                         let ext = &ext[pos..];
 
-                                        //FIXME(Jonathan): Here we are using a Vec<u8> when we probably want to
+                                        //FIXME: Here we are using a Vec<u8> when we probably want to
                                         //use a bytes::Bytes. Additionally we are copying the buffer
                                         //when moving or referencing is prerferrable.
                                         //The reason we're doing it this way is because there are
@@ -838,7 +829,7 @@ pub async fn upload_video(mut f: FormData) -> Result<String, CosmosError> {
                                         let ext = String::from(n);
                                         let ext = &ext[pos..];
 
-                                        //FIXME(Jonathan): Here we are using a Vec<u8> when we probably want to
+                                        //FIXME: Here we are using a Vec<u8> when we probably want to
                                         //use a bytes::Bytes. Additionally we are copying the buffer
                                         //when moving or referencing is prerferrable.
                                         //The reason we're doing it this way is because there are
@@ -942,10 +933,11 @@ pub struct CosmosSaga {
 
 enum Operation {
     Upsert {
-        old_document: Box<dyn erased_serde::Serialize + Send + Sync>,
+        old_document: Option<Box<dyn erased_serde::Serialize + Send + Sync>>,
         col_name: String,
         pk: PartitionKeys,
         etag: String,
+        document_id: String,
     },
     Insert {
         document_id: String,
@@ -974,8 +966,18 @@ impl Operation {
                 col_name,
                 pk,
                 etag,
+                document_id,
             } => {
-                upsert(col_name, pk, &old_document, Some(&etag)).await?;
+                // Old document might not exist in the case of this being an insert, in that case
+                // reverse by removing the new document
+                match old_document {
+                    Some(old_document) => {
+                        upsert(col_name, pk, &old_document, Some(&etag)).await?;
+                    }
+                    None => {
+                        delete(col_name, pk, document_id, Some(etag)).await?;
+                    }
+                }
             }
             Operation::Insert {
                 document_id,
@@ -1097,23 +1099,35 @@ impl CosmosSaga {
         document_id: I,
         etag: Option<&str>,
     ) -> Result<String, CosmosError> {
-        let (old_document, _): (D, _) =
-            get(collection_name.clone(), pk.clone(), document_id).await?;
-        let etag =
-            match upsert(collection_name.clone(), pk.clone(), document, etag)
-                .await
-            {
-                Ok(r) => r,
-                Err(e) => {
-                    self.abort().await?;
-                    return Err(e);
+        // Old document might not exist in the case of this upsert acting as an insert, in that
+        // case remember that this document did not exist
+        let old_document =
+            match get(collection_name.clone(), pk.clone(), document_id.to_string()).await {
+                Ok(r) => {
+                    let (r, _): (D, _) = r;
+                    let r: Box<dyn erased_serde::Serialize + Send + Sync> = Box::new(r);
+                    Some(r)
                 }
+                Err(e) => match e.kind {
+                    CosmosErrorKind::NotFound => None,
+                    _ => {
+                        return Err(e);
+                    }
+                },
             };
+        let etag = match upsert(collection_name.clone(), pk.clone(), document, etag).await {
+            Ok(r) => r,
+            Err(e) => {
+                self.abort().await?;
+                return Err(e);
+            }
+        };
         self.operation_stack.push(Operation::Upsert {
-            old_document: Box::new(old_document),
+            old_document,
             col_name: collection_name.to_string(),
             pk: pk.into(),
             etag: etag.clone(),
+            document_id: document_id.to_string(),
         });
         Ok(etag)
     }
@@ -1132,16 +1146,13 @@ impl CosmosSaga {
         document_id: I,
         etag: Option<&str>,
     ) -> Result<String, CosmosError> {
-        let etag =
-            match insert(collection_name.clone(), pk.clone(), document, etag)
-                .await
-            {
-                Ok(r) => r,
-                Err(e) => {
-                    self.abort().await?;
-                    return Err(e);
-                }
-            };
+        let etag = match insert(collection_name.clone(), pk.clone(), document, etag).await {
+            Ok(r) => r,
+            Err(e) => {
+                self.abort().await?;
+                return Err(e);
+            }
+        };
         self.operation_stack.push(Operation::Insert {
             document_id: document_id.to_string(),
             col_name: collection_name.to_string(),
